
#Import python libraries
import struct
from numpy import *


#Return the value of a register based off of its name
#See https://en.wikibooks.org/wiki/MIPS_Assembly/Register_File
def register(name):
  if name == '$zero': #Always zero
    return 0
  elif name == '$at': #Reserved for assembler
    return 1
  elif name == '$v0': #First and second return values, respectively
    return 2
  elif name == '$v1':
    return 3
  elif name == '$a0': #First four return arguents to functions
    return 4
  elif name == '$a1':
    return 5
  elif name == '$a2':
    return 6
  elif name == '$a3':
    return 7
  elif name == '$t0': #Temporary registers
    return 8
  elif name == '$t1':
    return 9
  elif name == '$t2':
    return 10
  elif name == '$t3':
    return 11
  elif name == '$t4':
    return 12
  elif name == '$t5':
    return 13
  elif name == '$t6':
    return 14
  elif name == '$t7':
    return 15
  elif name == '$s0': #Saved registers
    return 16
  elif name == '$s1':
    return 17
  elif name == '$s2':
    return 18
  elif name == '$s3':
    return 19
  elif name == '$s4':
    return 20
  elif name == '$s5':
    return 21
  elif name == '$s6':
    return 22
  elif name == '$s7':
    return 23
  elif name == '$t8': #More temporary registers
    return 24
  elif name == '$t9':
    return 25
  elif name == '$k0': #Reserved for kernel (operating system)
    return 26
  elif name == '$k1':
    return 27
  elif name == '$gp': #Global pointer
    return 28
  elif name == '$sp': #Stack pointer
    return 29
  elif name == '$fp': #Frame pointer
    return 30
  elif name == '$ra': #Return address
    return 31




#MIPS Assembler function, feed it a command as a string and it returns the hex
#This is not really meant to be a 100% fully functional, but to help it make it easier for me to hack the game
#For commands see https://en.wikibooks.org/wiki/MIPS_Assembly/Instruction_Formats
#For information about instruction types see https://web.archive.org/web/20180101004911if_/http://www.cs.umd.edu/class/spring2003/cmsc311/Notes/Mips/format.html
def assembler(asm):
  words = asm.lower().split(' ')
  n_words = len(words)
  command = words[0]
  opcode = 0x00
  funct = 0x00
  reg_s = 0x00
  reg_t = 0x00
  reg_d = 0x00
  imm = 0x0000
  address = 0x000000
  shift_amount = 0x00
  if command == 'nop': #Do nothing
    return(0x00000000)
  elif command == 'add': #Add
    itype = 'R'
    funct = 0x20
    reg_d = register(words[1])
    reg_s = register(words[2])
    reg_t = register(words[3])
  elif command == 'addi': #Add Immediate 
    itype = 'I'
    opcode = 0x08
    reg_t = register(words[1])
    reg_s = register(words[2])
    imm = int(words[3], 16)
  elif command == 'addiu': #Add Unsigned Immediate
    itype = 'I'
    opcode = 0x09
    reg_t = register(words[1])
    reg_s = register(words[2])
    imm = int(words[3], 16)
  elif command == 'addu': #Add Unsigned
    itype = 'R'
    funct = 0x21
    reg_d = register(words[1])
    reg_s = register(words[2])
    reg_t = register(words[3])
  elif command == 'and': #Bitwise AND
    itype = 'R'
    funct = 0x24
    reg_d = register(words[1])
    reg_s = register(words[2])
    reg_t = register(words[3])
  elif command == 'andi': #Bitwise AND Immediate
    itype = 'I'
    opcode = 0x0C
    reg_t = register(words[1])
    reg_s = register(words[2])
    imm = int(words[3], 16)
  elif command == 'beq': #Branch if Equal
    itype = 'I'
    opcode = 0x04
    reg_s = register(words[1])
    reg_t = register(words[2])
    imm = int(words[3], 16)
  elif command == 'blez': #Branch if Less Than or Equal to Zero
    itype = 'I'
    opcode = 0x06
    reg_s = register(words[1])
    imm = int(words[2], 16)
  elif command == 'bne': #Branch if Not Equal
    itype = 'I'
    opcode = 0x05
    reg_s = register(words[1])
    reg_t = register(words[2])
    imm = int(words[3], 16)
  elif command == 'bgtz': #Branch on Greater Than Zero
    itype = 'I'
    opcode = 0x07
    reg_s = register(words[1])
    imm = int(words[2], 16)
  elif command == 'div': #Divide
    itype = 'R'
    funct = 0x1A
    reg_s = register(words[1])
    reg_t = register(words[2])
  elif command == 'divu': #Unsigned Divide
    itype = 'R'
    funct = 0x1B
    reg_s = register(words[1])
    reg_t = register(words[2])
  elif command == 'j': #Jump to Address
    itype = 'J'
    opcode = 0x02
    address = int((int(words[1], 16) - 0x80000000)/4)
  elif command == 'jal': #Jump and Link
    itype = 'J'
    opcode = 0x03
    address = int((int(words[1], 16) - 0x80000000)/4)
  elif command == 'jr': #Jump to Address in Register
    itype = 'R'
    funct = 0x08
    reg_s = register(words[1])
  elif command == 'lb': #Load Byte
    itype = 'I'
    opcode = 0x20
    reg_t = register(words[1])
    imm = int(words[2], 16)
    reg_s = register(words[3])
  elif command == 'lbu': #Load Byte Unsigned
    itype = 'I'
    opcode = 0x24
    reg_t = register(words[1])
    imm = int(words[2], 16)
    reg_s = register(words[3])
  elif command == 'lhu': #Load Halfword Unsigned
    itype = 'I'
    opcode = 0x25
    reg_t = register(words[1])
    imm = int(words[2], 16)
    reg_s = register(words[3])
  elif command == 'lh': #Load Halfword 
    itype = 'I'
    opcode = 0x21
    reg_t = register(words[1])
    imm = int(words[2], 16)
    reg_s = register(words[3])
  elif command == 'lui': #Load Upper Immediate
    itype = 'I'
    opcode = 0x0F
    reg_t = register(words[1])
    imm = int(words[2], 16)
  elif command == 'lw': #Load Word
    itype = 'I'
    opcode = 0x23
    reg_t = register(words[1])
    imm = int(words[2], 16)
    reg_s = register(words[3])
  elif command == 'mfhi': #Move from HI Register
    itype = 'R'
    funct = 0x10
    reg_d = register(words[1])
  elif command == 'mthi': #Move to HI Register
    itype = 'R'
    funct = 0x11
    reg_s = register(words[1])
  elif command == 'mflo': #Move from LO Register
    itype = 'R'
    funct = 0x12
    reg_d = register(words[1])
  elif command == 'mtlo': #Move to LO Register
    itype = 'R'
    funct = 0x13
    reg_s = register(words[1])
  elif command == 'mfc0': #Move from Coprocessor 0
    itype = 'R'
    opcode = 0x10
  elif command == 'mult': #Multiply
    itype = 'R'
    funct = 0x18
    reg_s = register(words[1])
    reg_t = register(words[2])
  elif command == 'multu': #Unsigned Multiply
    itype = 'R'
    funct = 0x19
    reg_s = register(words[1])
    reg_t = register(words[2])
  elif command == 'nor': #Bitwise NOR (NOT-OR)
    itype = 'R'
    funct = 0x27
    reg_d = register(words[1])
    reg_s = register(words[2])
    reg_t = register(words[3])
  elif command == 'xor': #Bitwise XOR (Exclusive-OR)
    itype = 'R'
    funct = 0x26
    reg_d = register(words[1])
    reg_s = register(words[2])
    reg_t = register(words[3])
  elif command == 'or': #Bitwise OR
    itype = 'R'
    funct = 0x25
    reg_d = register(words[1])
    reg_s = register(words[2])
    reg_t = register(words[3])
  elif command == 'ori': #Bitwise OR Immediate
    itype = 'I'
    opcode = 0x0D
    reg_t = register(words[1])
    reg_s = register(words[2])
    imm = int(words[3], 16)
  elif command == 'sb': #Store Byte
    itype = 'I'
    opcode = 0x28
    reg_t = register(words[1])
    imm = int(words[2], 16)
    reg_s = register(words[3])
  elif command == 'sh': #Store Halfword
    itype = 'I'
    opcode = 0x29
    reg_t = register(words[1])
    imm = int(words[2], 16)
    reg_s = register(words[3])
  elif command == 'slt': #Set to 1 if Less Than
    itype = 'R'
    funct = 0x2A
    reg_d = register(words[1])
    reg_s = register(words[2])
    reg_t = register(words[3])
  elif command == 'slti': #Set to 1 if Less Than Immediate
    itype = 'I'
    opcode = 0x0A
    reg_t = register(words[1])
    reg_s = register(words[2])
    imm = int(words[3], 16)
  elif command == 'sltiu': #Set to 1 if Less Than Unsigned Immediate
    itype = 'I'
    opcode = 0x0B
    reg_t = register(words[1])
    reg_s = register(words[2])
    imm = int(words[3], 16)
  elif command == 'sltu': #Set to 1 if Less Than Unsigned
    itype = 'R'
    funct = 0x2B
    reg_d = register(words[1])
    reg_s = register(words[2])
    reg_t = register(words[3])
  elif command == 'sll': #Logical Shift Left
    itype = 'R'
    opcode = 0x00
    reg_d = register(words[1])
    reg_t = register(words[2])
    shift_amount = int(words[3], 16)
  elif command == 'srl': #Logical Shift Right (0-extended)
    itype = 'R'
    funct = 0x02
    reg_d = register(words[1])
    reg_t = register(words[2])
    shift_amount = int(words[3], 16)
  elif command == 'sra': #Arithmetic Shift Right (sign-extended)
    itype = 'R'
    funct = 0x03
    reg_d = register(words[1])
    reg_t = register(words[2])
    shift_amount = int(words[3], 16)
  elif command == 'sub': #Subtract
    itype = 'R'
    funct = 0x22
    reg_d = register(words[1])
    reg_s = register(words[2])
    reg_t = register(words[3])
  elif command == 'subu': #Unsigned Subtract
    itype = 'R'
    funct = 0x23
    reg_d = register(words[1])
    reg_s = register(words[2])
    reg_t = register(words[3])
  elif command == 'sw': #Store Word
    itype = 'I'
    opcode = 0x2B
    reg_t = register(words[1])
    imm = int(words[2], 16)
    reg_s = register(words[3])
  if itype == 'R':
    return(opcode*2**26 + reg_s*2**21 + reg_t*2**16 + reg_d*2**11 + shift_amount*2**6 + funct)
  elif itype == 'I':
    return(opcode*2**26 + reg_s*2**21 + reg_t*2**16 + imm)
  elif itype == 'J':
    return(opcode*2**26 + address)
  else:
    print('ERROR: Something went wrong with '+asm)
  return(result)

#Convert asssembly (or hex) to gameshark code
def asm_to_gameshark(ram_address, asm):
  if size(asm) == 1: #Handle single lines
    asm = [asm]
  gameshark_code = [] #List to hold output
  current_byte = 0 #Track memory address where GS code is going
  for asm_line in asm: #Loop through each ASM instruction
    if type(asm_line) == str: #If ASM is a string
      asm_converted_to_hex = assembler(asm_line) #Run the assembler to convert asm to hex
    elif type(asm_line) == int: #If ASM already is in hex
      asm_converted_to_hex = asm_line #Just use the existing hex
    asm_converted_to_hex_str = hex(asm_converted_to_hex)[2:].upper() #Convert hex for ASM instruction into a string
    gameshark_code.append('81'+hex(ram_address+current_byte)[4:].upper()+' '+asm_converted_to_hex_str[0:4]) #Write first 16 bits of 32 bit ASM instruction as an 81 type gs code
    current_byte = current_byte + 2 #Increment current byte
    gameshark_code.append('81'+hex(ram_address+current_byte)[4:].upper()+' '+asm_converted_to_hex_str[4:8]) #Write last 16 bits of 32 bit ASM instruction as an 81 type gs code
    current_byte = current_byte + 2
  return gameshark_code



#Functions to insert instrucitons or data into the rom, shamelessly ripped from Micro500's micro mountain python scirpt
def mem_read_u32_be(addr, data):
  return struct.unpack(">I",data[addr:addr+4])[0]
  
def mem_read_u16_be(addr, data):
  return struct.unpack(">H",data[addr:addr+2])[0]
  
def u16_to_le(data):
    return struct.pack("<H",data)

def u16_to_be(data):
    return struct.pack(">H",data)

def s16_to_be(data):
    return struct.pack(">h",data)

def u32_to_be(data):
    return struct.pack(">I",data)
  
def mem_fill(data, start_addr, end_addr, value):
  return data[0:start_addr] + (bytes([value] * (end_addr - start_addr))) + data[end_addr:len(data)]
  
def mem_write_u32_be(addr, value, data):
  return data[0:addr] + u32_to_be(value) + data[(addr+4):len(data)]

def mem_write_u16_be(addr, value, data):
  return data[0:addr] + u16_to_be(value) + data[(addr+2):len(data)]
  
def mem_write_u8_be(addr, value, data):
  return data[0:addr] + u8_to_be(value) + data[(addr+1):len(data)]

def mem_write_16(addr, value, data):
  return data[0:addr] + value + data[(addr+2):len(data)]

# #Convert Gameshark like script into ASM (AKA compile)
# def script_to_asm(script):
#   line = 1
#   script_to_asm = []
#   for script_line in script:
#     words = script_line.split().lower()
#     if words[1] == 'if':
#       #do stuff
#     elif words[1] == 'while':
#       #do stuff
#     elif words[1] == 'for':
#       #do stuff
 



def gameshark_to_hex(gameshark_codes):
    line = 1
    gameshark_hex = []
    for gameshark_code in gameshark_codes: #Build the code handler
      code_type = gameshark_code[0:2]
      mem_address = gameshark_code[2:8]
      mem_address_suffix = mem_address[2:6]
      if int(mem_address_suffix, 16) >= 0x8000:
        mem_address_prefix = hex(0x8001 + int(mem_address[0:2], 16))[2:].upper()
      else:
        mem_address_prefix = hex(0x8000 + int(mem_address[0:2], 16))[2:].upper()
      #print('Trying to add '+gameshark_code)
      if code_type == '80' or code_type == 'D0' or code_type == 'D2':
        code_value = gameshark_code[11:13]
      elif code_type == '81' or code_type == 'D1' or code_type == 'D3':
        code_value = gameshark_code[9:13]
      if code_type == '80':
        gameshark_hex.append('0x3C1A'+mem_address_prefix) #LUI $K0 0xXXXX
        gameshark_hex.append('0x241B00'+code_value) #ADDIU $K1 $ZERO, 0x00XX
        gameshark_hex.append('0xA35B'+mem_address_suffix) #SB $K1 0xXXXX ($K0)
        #gameshark_hex.append('0x00000000') #Do nothing
        #print('Added '+gameshark_code)
      elif code_type == '81':
        gameshark_hex.append('0x3C1A'+mem_address_prefix) #LUI $K0 0xXXXX
        gameshark_hex.append('0x241B'+code_value) #ADDIU $K1 $ZERO 0xXXXX
        gameshark_hex.append('0xA75B'+mem_address_suffix) #SH $K1 0xXXXX ($K0)
        #gameshark_hex.append('0x00000000') #Do nothing
      elif code_type == 'D0':
        gameshark_hex.append('0x3C1A'+mem_address_prefix) #LUI $K0 0xXXXX
        gameshark_hex.append('0x935A'+mem_address_suffix) #LBU $K0 0x0000 ($K0)
        gameshark_hex.append('0x241B00'+code_value) #ADDIU $K1 $ZERO 0x00XX
        gameshark_hex.append('0x175B0003') #BNE $K0 $K1 0x0004
      elif code_type == 'D1':
        gameshark_hex.append('0x3C1A'+mem_address_prefix) #LUI $K0 0xXXXX
        gameshark_hex.append('0x975A'+mem_address_suffix) #LHU $K0 0xXXXX ($K0)
        gameshark_hex.append('0x241B'+code_value) #ADDIU $K1 $ZERO 0xXXXX
        gameshark_hex.append('0x175B0003') #BNE $K0 $K1 0x0004
      elif code_type == 'D2':
        gameshark_hex.append('0x3C1A'+mem_address_prefix) #LUI $K0 0xXXXX
        gameshark_hex.append('0x935A'+mem_address_suffix) #LBU $K0 0x0000 ($K0)
        gameshark_hex.append('0x241B00'+code_value) #ADDIU $K1 $ZERO 0x00??
        gameshark_hex.append('0x135B0003') #BEQ $K0 $K1 0x0004
      elif code_type == 'D3':
        gameshark_hex.append('0x3C1A'+mem_address_prefix) #LUI $K0 0xXXXX
        gameshark_hex.append('0x975A'+mem_address_suffix) #LHU $K0 0x0000 ($K0)
        gameshark_hex.append('0x241B'+code_value) #ADDIU $K1, $ZERO, 0xXXXX
        gameshark_hex.append('0x135B0003')#BEQ $K0, $K1, 0x0004
      else:
        print('WARNING:  Line '+str(line)+ '('+gameshark_code+') in Gameshark code list is invalid.')
      line = line + 1
    gameshark_hex.append('0x00000000')
    #print('Inserting the following hex from Gameshark codes into the ROM:')
    for i in range(len(gameshark_hex)): #Convert strings to ints
      #print(gameshark_hex[i])
      gameshark_hex[i] = int(gameshark_hex[i], 16)
    return gameshark_hex


# def gameshark_to_hex(gameshark_codes):
#     line = 1
#     gameshark_hex = []
#     for gameshark_code in gameshark_codes: #Build the code handler
#       code_type = gameshark_code[0:2]
#       mem_address = gameshark_code[2:8]
#       #print('Trying to add '+gameshark_code)
#       if code_type == '80' or code_type == 'D0' or code_type == 'D2':
#         code_value = gameshark_code[11:13]
#       elif code_type == '81' or code_type == 'D1' or code_type == 'D3':
#         code_value = gameshark_code[9:13]
#       if code_type == '80':
#         gameshark_hex.append('0x3C1A80'+mem_address[0:2]) #LUI $K0 0x80XX
#         gameshark_hex.append('0x375A'+mem_address[2:6]) #ORI $K0 $K0 0xXXXX
#         gameshark_hex.append('0x241B00'+code_value) #ADDIU $K1 $ZERO, 0x00XX
#         gameshark_hex.append('0xA35B0000') #SB $K1 0x0000 ($K0)
#         gameshark_hex.append('0x00000000') #Do nothing
#         gameshark_hex.append('0x00000000') #Do nothing
#         print('Added '+gameshark_code)
#       elif code_type == '81':
#         gameshark_hex.append('0x3C1A80'+mem_address[0:2]) #LUI $K0, 0x80XX
#         gameshark_hex.append('0x375A'+mem_address[2:6]) #ORI $K0 $K0 0xXXXX
#         gameshark_hex.append('0x241B'+code_value) #ADDIU $K1 $ZERO 0xXXXX
#         gameshark_hex.append('0xA75B0000') #SH $K1 0x0000 ($K0)
#         gameshark_hex.append('0x00000000') #Do nothing
#         gameshark_hex.append('0x00000000') #Do nothing
#       elif code_type == 'D0':
#         gameshark_hex.append('0x3C1A80'+mem_address[0:2]) #LUI $K0 0x80XX
#         gameshark_hex.append('0x375A'+mem_address[2:6]) #ORI $K0 $K0 0xXXXX
#         gameshark_hex.append('0x835A0000') #LB $K0 0x0000 ($K0)
#         gameshark_hex.append('0x241B00'+code_value) #ADDIU $K1 $ZERO 0x00XX
#         gameshark_hex.append('0x175B0007') #BNE $K0 $K1 0x0007
#         gameshark_hex.append('0x00000000') #Do nothing
#       elif code_type == 'D1':
#         gameshark_hex.append('0x3C1A80'+mem_address[0:2]) #LUI $K0 0x80XX
#         gameshark_hex.append('0x375A'+mem_address[2:6]) #ORI $K0 $K0 0xXXXX
#         gameshark_hex.append('0x875A0000') #LH $K0 0x0000 ($K0)
#         gameshark_hex.append('0x241B'+code_value) #ADDIU $K1 $ZERO 0xXXXX
#         gameshark_hex.append('0x175B0007') #BNE $K0 $K1 0x0007
#         gameshark_hex.append('0x00000000') #Do nothing
#       elif code_type == 'D2':
#         gameshark_hex.append('0x3C1A80'+mem_address[0:2]) #LUI $K0 0x80XX
#         gameshark_hex.append('0x375A'+mem_address[2:6]) #ORI $K0 $K0 0xXXXX
#         gameshark_hex.append('0x835A0000') #LB $K0 0x0000 ($K0)
#         gameshark_hex.append('0x241B00'+code_value) #ADDIU $K1 $ZERO 0x00??
#         gameshark_hex.append('0x135B0007') #BEQ $K0 $K1 0x0007
#         gameshark_hex.append('0x00000000') #Do nothing
#       elif code_type == 'D3':
#         gameshark_hex.append('0x3C1A80'+mem_address[0:2]) #LUI $K0 0x80XX
#         gameshark_hex.append('0x375A'+mem_address[2:6]) #ORI $K0 $K0 0xXXXX
#         gameshark_hex.append('0x875A0000') #LH $K0 0x0000 ($K0)
#         gameshark_hex.append('0x241B'+code_value) #ADDIU $K1, $ZERO, 0xXXXX
#         gameshark_hex.append('0x135B0007')#BEQ $K0, $K1, 0x0007
#         gameshark_hex.append('0x00000000') #Do nothing
#       else:
#         print('WARNING:  Line '+str(line)+ '('+gameshark_code+') in Gameshark code list is invalid.')
#       line = line + 1
#     for i in range(len(gameshark_hex)): #Convert strings to ints
#       gameshark_hex[i] = int(gameshark_hex[i], 16)
#     return gameshark_hex





# def put_hex_code_into_rom(code, memlocation, romdata):
#   n_lines = len(code)
#   for i in range(n_lines):
#     romdata = mem_write_u32_be(memlocation, code[i], romdata)
#     memlocation += 4
#   return romdata



#Class handles a DMA codeblock
class dma_codeblock:
  def __init__(self):
    self.code = []
    self.romaddr = 0x0000000 #Location of code source in rom
    self.ramaddr = 0x0000000 #Destination of code in ram
    self.hookaddr = 0x00000000
  def size(self): #Get size of code block
    #return int(len(self.code))
    return int(len(self.code)*4)




#Class handles everything for patching a rom
class rom:
  def __init__(self, file_name): 
    self.load(file_name)
    #self.offset = -0xC00 #Offset between RAM and ROM for first MB of code loaded RAM = ROM + offset (for MK64, it is -0xC00)
    self.rom_to_ram = mem_read_u32_be(0x8, self.data) - 0x1000 #Grab converstion of RAM to ROM (negative to go ROM to RAM) from ram address of the bootcode at 0x8 in the rom
    self.ram_to_rom = - self.rom_to_ram
    self.dma_copy_addr = 0x80001158 #Address for dma_copy function
    #self.dma_copy_addr = 0x800D3520 #Address for the byte copy function
    #self.dma_copy_run_codeblock_addr = 0x58084  #somewhere in the debug to run the DMA copy code
    self.dma_copy_run_codeblock_addr = 0xE9E80 #Somewhere in PAL VI code that goes unused in NTSC (seems to be the best so far) to run the DMA copy code
    #self.dma_copy_run_hook_addr = 0x4704 #Try hooking the DMA funciton here
    #self.dma_copy_run_hook_addr = 0x14A4 #Try hooking the DMA funciton here
    #self.dma_copy_run_hook_addr = 0x1E64 #Try hooking the DMA funciton here
    #self.dma_copy_run_hook_addr = 0x29F0 #Try hooking the DMA funciton here
    self.dma_copy_run_hook_addr = 0x17EC #Try hooking the DMA funciton here, found by Rain
    #self.dma_copy_run_hook_addr = 0xBFFFFFF0 #Try hooking the DMA funciton here
    self.dma_codeblocks = [] #List to store DMA code blocks
  def load(self, file_name): #Load rom data from file (.z64)
    rom_file = open(file_name, 'rb')
    self.data = rom_file.read()
    rom_file.close()
  def save(self, file_name, checksum=True): #Save a patched rom
    if len(self.dma_codeblocks) > 0: #Run if any DMA codeblocks exist
      self.put_dma_codeblocks_into_rom()
    if checksum:
      self.fix_checksum() #Fix checksum before saving patched rom (important this is the last step before saving the rom)
    patched_rom_file = open(file_name, 'wb')
    patched_rom_file.write(self.data)
    patched_rom_file.close()
  def insert_hex(self, addr, hexcode): #Insert 32 bit MIPs instruction(s) or data in hex format into rom
      if type(hexcode) == int:  #If there is only one line
        self.data =  mem_write_u32_be(addr, hexcode, self.data)#Insert a single 32 bit value into ram
      else: #If there are multiple lines
        for line in hexcode: #Loop through each line
          self.data =  mem_write_u32_be(addr, line, self.data)#Insert a single 32 bit value into ram
          addr += 4 #Go to next line
  def insert_code(self, addr, code):
      try:
        if type(code) == int:
          self.data =  mem_write_u32_be(addr, code, self.data)#Insert a single 32 bit value into ram
        elif type(code) == str:  #If there is only one line
          self.data =  mem_write_u32_be(addr, assembler(code), self.data)#Insert a single 32 bit value into ram
      except:
        print('INSERT CODE ERROR: ', code)
      else: #If there are multiple lines
        line_number = 0
        for line in code: #Loop through each line
          line_number += 1
          try:
            if type(line) == int:
              self.data =  mem_write_u32_be(addr, line, self.data)#Insert a single 32 bit value into ram
              #print('Inserting ', '0x'+format(line, '08x').upper(), ' into rom at ', '0x'+format(addr, '08x').upper())
            elif type(line) == str:
              self.data =  mem_write_u32_be(addr, assembler(line), self.data)#Insert a single 32 bit value into ram
              #print('Inserting ', line, ';', '0x'+format(assembler(line), '08x').upper(), ' into rom at ', '0x'+format(addr, '08x').upper())
          except:
            print('INSERT CODE ERROR ON LINE '+str(line_number)+': ', line)
          addr += 4 #Go to next line  
  def insert_codeblock(self, hook_addr, codeblock_addr, codeblock, jump_addr=0, offset=0, ignore_offset=False): #Insert a block of custom code with a hook (and custom jump back if you want)
    #if offset == 0 and not ignore_offset: #Use game's default offset for first 1 MB if no offset is provided
    #  offset = self.offset
    hook_instructions = [0x08000000 + int((codeblock_addr + self.rom_to_ram - 0x80000000)/4),0x00000000]
    self.insert_code(hook_addr, hook_instructions)
    if jump_addr == 0: #If no jump address is specified, just jump back to right after the hook
      jump_back = [0x08000000 + 1 + int((hook_addr + self.rom_to_ram - 0x80000000)/4), 0x00000000]
    else: #Else use the specified jump address
      jump_back = [0x08000000 + int((jump_addr + self.rom_to_ram - 0x80000000)/4), 0x00000000]
    self.insert_code(codeblock_addr, codeblock + jump_back)    
  def insert_gameshark(self, hook_addr, codehandler_addr, gameshark_codes): #Insert a gameshark code handler engine along with desired codes
    gameshark_hex = gameshark_to_hex(gameshark_codes)
    self.insert_codeblock(hook_addr, codehandler_addr, gameshark_hex) #Insert the gameshark code handler block into the rom with a hook
  def insert_dma_codeblock(self, hookaddr, romaddr, ramaddr, code): #Insert a new DMA codeblock into the rom
    new_dma_codeblock = dma_codeblock() #Create new DMA codeblock obj
    new_dma_codeblock.romaddr = romaddr #Set the rom source address
    new_dma_codeblock.ramaddr = ramaddr #Set the ram destination address
    new_dma_codeblock.hookaddr = hookaddr
    jump_back = [0x08000000 + 1 + int((hookaddr + self.rom_to_ram - 0x80000000)/4), 0x00000000] #Insert jump back to hook
    new_dma_codeblock.code = code + jump_back #Set the codeblock
    self.dma_codeblocks.append(new_dma_codeblock) #Stack the DMA code blocks for later inserting into the rom
  def put_dma_codeblocks_into_rom(self): #Insert snippit of code that will load a block of code from rom into ram (see how it is done in https://github.com/abitalive/MarioKart64/blob/master/Patches/Multiplayer_Hack/mk64_multiplayer_hack.asm)
    run_dmacopy_asm = [
      #'lui $a0 0x8000', #Set memory to use expansion pak and Store $ra in stack
      #'ori $a1 $zero 0x0080',
      #'sh $a1 0x0013 $a0',
      #'sub $sp $sp 0x4',  #Store $ra in stack
      #'sw $ra 0x0000 $sp'
      ] 
    for current_dma_codeblock in self.dma_codeblocks:
      hook_instructions = [0x08000000 + int((current_dma_codeblock.ramaddr-0x80000000)/4),0x00000000] #Set up hook to DMA codeblock
      self.insert_code(current_dma_codeblock.hookaddr, hook_instructions) #Insert hook into rom
      self.insert_code(current_dma_codeblock.romaddr, current_dma_codeblock.code) #Insert the codeblock for running DMA copy on into the rom
      romaddr_hex_str = format(current_dma_codeblock.romaddr, '08x')  #The rest of this code creates the code to copy from rom into ram using dma copy
      ramaddr_hex_str = format(current_dma_codeblock.ramaddr, '08x')
      size_hex_str = format(current_dma_codeblock.size(), '08x')
      run_dmacopy_asm +=  [
        'ADDIU $SP $SP 0xFFE0', #Store registers in stack, ADDIU sp, sp, -0x20
        'sw $A0 0x0010 $SP',
        'sw $A1 0x0014 $SP',
        'sw $A2 0x0018 $SP',
        'lui $a0 0x'+ramaddr_hex_str[0:4],  #Load RAM destination
        'ori $a0 $a0 0x'+ramaddr_hex_str[4:8],
        'lui $a1 0x'+romaddr_hex_str[0:4], #Load ROM source
        'ori $a1 $a1 0x'+romaddr_hex_str[4:8],
        'lui $a2 0x'+size_hex_str[0:4], #Load size
        'ori $a2 $a2 0x'+size_hex_str[4:8],
        'jal '+hex(int(self.dma_copy_addr)), #JAL DmaCopy
        'nop',
        'lw $A0 0x0010 $SP',  #Load registers from stack
        'lw $A1 0x0014 $SP',
        'lw $A2 0x0018 $SP',
        'ADDIU $SP $SP 0x20',
        'LUI $T6 0x8030', #Run these two lines the hook replaced
        'LUI $AT 0x1FFF',
      ]
    self.insert_codeblock(self.dma_copy_run_hook_addr, self.dma_copy_run_codeblock_addr, run_dmacopy_asm, jump_addr=0x12AC+0xC00, ignore_offset=False)
  def fix_checksum(self):   #Fix checksum for rom, shamelessly ripped from Micro500's micro mountain python scirpt
      # Ported from https://github.com/queueRAM/sm64tools/blob/master/n64cksum.c
      # by dkosmari https://gist.github.com/dkosmari/ee7bb471ea12c21b008d0ecffebd6384
      #local t0, t1, t2, t3, t4, t5, t6, t7, t8, t9
      #local s0, s6
      #local a0, a1, a2, a3, at
      #local lo
      #local v0, v1
      #local ra
      mask32 = 0xFFFFFFFF
      # derived from the SM64 boot code
      s6 = 0x3f
      a0 = 0x1000                     # 59c:   8d640008    lw a0,8(t3)
      a1 = s6                         # 5a0:   02c02825    move  a1,s6
      at = 0x5d588b65                 # 5a4:   3c015d58    lui   at,0x5d58
                                      # 5a8:   34218b65    ori   at,at,0x8b65
      lo = (a1 * at) & mask32         # 5ac:   00a10019    multu a1,at    16 F8CA 4DDB
      ra = 0x100000                   # 5bc:  3c1f0010    lui   ra,0x10
      v1 = 0                          # 5c0:  00001825    move  v1,zero
      t0 = 0                          # 5c4:  00004025    move  t0,zero
      t1 = a0                         # 5c8:  00804825    move  t1,a0
      t5 = 32                         # 5cc:  240d0020    li t5,32
      v0 = lo                         # 5d0:  00001012    mflo  v0
      v0 = (v0 + 1) & mask32          # 5d4:  24420001    addiu v0,v0,1
      a3 = v0                         # 5d8:  00403825    move  a3,v0
      t2 = v0                         # 5dc:  00405025    move  t2,v0
      t3 = v0                         # 5e0:  00405825    move  t3,v0
      s0 = v0                         # 5e4:  00408025    move  s0,v0
      a2 = v0                         # 5e8:  00403025    move  a2,v0
      t4 = v0                         # 5ec:  00406025    move  t4,v0
      while t0 != ra:
          v0 = mem_read_u32_be(t1, self.data)   # 5f0: 8d220000    lw v0,0(t1)
          v1 = (a3 + v0) & mask32     # 5f4: 00e21821    addu  v1,a3,v0
          at = v1 < a3                # 5f8: 0067082b    sltu  at,v1,a3
          a1 = v1                     # 600: 00602825    move  a1,v1 branch delay slot
          if at:                      # 5fc: 10200002    beqz  at,0x608
              t2 = (t2 + 1) & mask32  # 604: 254a0001    addiu t2,t2,1

          v1 = v0 & 0x1F              # 608: 3043001f    andi  v1,v0,0x1f
          t7 = (t5 - v1) & mask32     # 60c: 01a37823    subu  t7,t5,v1
          t8 = v0 >> t7               # 610: 01e2c006    srlv  t8,v0,t7
          t6 = (v0 << v1) & mask32    # 614: 00627004    sllv  t6,v0,v1
          a0 = t6 | t8                # 618: 01d82025    or a0,t6,t8
          at = a2 < v0                # 61c: 00c2082b    sltu  at,a2,v0
          a3 = a1                     # 620: 00a03825    move  a3,a1
          t3 = (t3 ^ v0) & mask32     # 624: 01625826    xor   t3,t3,v0
          s0 = (s0 + a0) & mask32     # 62c: 02048021    addu  s0,s0,a0 branch delay slot
          if at:                      # 628: 10200004    beqz  at,0x63c
              t9 = (a3 ^ v0) & mask32 # 630: 00e2c826    xor   t9,a3,v0
                                      # 634: 10000002    b  0x640
              a2 = (a2 ^ t9) & mask32 # 638: 03263026    xor   a2,t9,a2 branch delay
          else:
              a2 = (a2 ^ a0) & mask32 # 63c: 00c43026    xor   a2,a2,a0

          t0 += 4                     # 640: 25080004    addiu t0,t0,4
          t7 = (v0 ^ s0) & mask32     # 644: 00507826    xor   t7,v0,s0
          t1 += 4                     # 648: 25290004    addiu t1,t1,4
          t4 = (t4 + t7) & mask32     # 650: 01ec6021    addu  t4,t7,t4 branch delay
                                      # 64c: 151fffe8    bne   t0,ra,0x5f0
      t6 = (a3 ^ t2) & mask32         # 654: 00ea7026    xor   t6,a3,t2
      a3 = (t6 ^ t3) & mask32         # 658: 01cb3826    xor   a3,t6,t3
      t8 = (s0 ^ a2) & mask32         # 65c: 0206c026    xor   t8,s0,a2
      s0 = (t8 ^ t4) & mask32         # 660: 030c8026    xor   s0,t8,t4
      self.data = mem_write_u32_be(0x10, a3, self.data)
      self.data = mem_write_u32_be(0x14, s0, self.data)
  def disable_internal_aa(self): #Loops through entire rom and disables internal AA from the F3DEX microcode (see https://www.assembler-games.com/threads/is-it-possible-to-disable-anti-aliasing-in-n64-games-via-gameshark-cheats.59916/page-3#post-860932)
    rom_size = len(self.data)
    for addr in range(0x0, rom_size, 0x4): #Loop through entire rom
      instruction = mem_read_u32_be(addr, self.data)
      if instruction == 0xB900031D:
        print('------------')
        next_instruction =  mem_read_u32_be(addr + 0x4, self.data)
        # if next_instruction & (1 << 3): #If 3rd bit is set (internal AA is on)
        #   print('Changed ', hex(next_instruction))
        #   next_instruction = next_instruction - 0x8 #Turn internal AA off
        #   print('into ', hex(next_instruction))
        #   self.data = mem_write_u32_be(addr + 0x4, next_instruction, self.data)
        if  (next_instruction & (1 << 4)): #If 3rd bit is set (internal AA is on)
          print('Changed ', hex(next_instruction))
          next_instruction = next_instruction - 0x10   #do some random ass thing
          print('into ', hex(next_instruction))
          self.data = mem_write_u32_be(addr + 0x4, next_instruction, self.data)
  def get_address(self, instruction): #Grab and return the rom address of an instruction
    if not type(instruction) == int: #Convert instruction to hex if not done already
      intruction = assembler(instruction)
    rom_size = len(self.data)
    for addr in range(0x0, rom_size, 0x4): #Loop through entire rom
      if instruction == mem_read_u32_be(addr, self.data): #If the instruction matches, return the address
        return(addr)

#Class returns memory locations for the player structure 
#  
#  [NEEDS A LOT OF WORK TO ENSURE I HAVE TH RIGHT RAM ADDESSES, SOME SEEM OFF AFTER U5]
#
# struct mk64_player
# {
#    float unkn;
#    u32 u1;
#    u32 u2;
#    u32 u3;
#    u32 u4;

#    float x, y, z;
#    u16 u5;
#    u16 rotation;

#    float unkn;

#    float x_accel, y_accel, z_accel;
# };

class player:
  def __init__(self, offset):
    base = 0x800F6990 + offset
    self.unkn = base
    self.u1 = base + 4
    self.u2 = base + 8
    self.u3 = base + 12
    self.u4 = base + 16
    self.x = base + 20
    self.y = base + 24
    self.z = base + 28
    self.u5 = base + 30
    self.rotation = base + 32
    self.unkn = base + 34
    self.x_accel = base + 38
    self.y_accel = base + 42
    self.z_accel = base + 46

#Generate a list storing the memory locations for all 8 player structures
offset = 3544
players = []
for i in range(8):
  players.append(player(offset*i))


#Class for handling track segment offsets



#class that stores course table info for the following structure: 


# Offset  Length  Description
# 0x00  8 MIO0 block containing display lists, item box placements, path data, and ??; decoded to segment 06
# 0x08  8 Copied to segment 0F - MIO0 block of course vertex data (gfx & collisions); decoded to segment 04, followed by packed display list, decoded to segment 07
# 0x10  8 Course texture MIO0 block references (segment 05), segment 06 display list references
# 0x18  4 Segment to load vertex data and packed DL pre-decompression/unpacking ?
# 0x1C  4 ? (gets multiplied by 24 at some point)
# 0x20  4 Segment offset of packed display list (unpacked to segment 7)
# 0x24  4 Segment 7 final DL command offset (follows the DL commands that reference all other DLs in the segment)
# 0x28  4 Segment to load texture list and DL jump table?
# 0x2C  2 ? 0000 or 0001
# 0x2E  2 struct padding, unused

#See https://hack64.net/wiki/doku.php?id=mario_kart_64:course_data_table#mario_kart_64_course_data_table


# class course_table_entry:
#   def __init__(self, address,  offset):
#     self.address = address
#     self.offset = offset

# class course_table_range:
#   def __init__(self, start, end, offset):
#     self.start = start
#     self.end = end
#     self.offset = offset



# Offset  Length  Description
# 0x00  8 MIO0 block containing display lists, item box placements, path data, and ??; decoded to segment 06
# 0x08  8 Copied to segment 0F - MIO0 block of course vertex data (gfx & collisions); decoded to segment 04, followed by packed display list, decoded to segment 07
# 0x10  8 Course texture MIO0 block references (segment 05), segment 06 display list references
# 0x18  4 Segment to load vertex data and packed DL pre-decompression/unpacking ?
# 0x1C  4 ? (gets multiplied by 24 at some point)
# 0x20  4 Segment offset of packed display list (unpacked to segment 7)
# 0x24  4 Segment 7 final DL command offset (follows the DL commands that reference all other DLs in the segment)
# 0x28  4 Segment to load texture list and DL jump table?
# 0x2C  2 ? 0000 or 0001
# 0x2E  2 struct padding, unused

#See https://hack64.net/wiki/doku.php?id=mario_kart_64:course_data_table#mario_kart_64_course_data_table
course_table = [[0x008284D0, 0x0082B620, 0x0088FA10, 0x0089B510, 0x0088CD70, 0x0088D070, 0x0F000000, 0x0000167D, 0x0F0096F4, 0x00006930, 0x09000000, 0x00000000], #Mario Raceway ID=0x0000
  ]


def generate_course_table_entry_asm(course_table_row, course_id): #Generates 38 (0x26) instrucitons 
  address = 0x802B8D80 + 0x30 * course_id
  asm = []
  address_str = format(address, '08x')
  asm.append('lui $k0 0x'+address_str[0:4])
  asm.append('ori $k0 $k0 '+address_str[4:8])
  counter = 0x00
  for course_table_col in course_table_row:
    hex_str = format(course_table_col, '08x')
    asm.append('lui $k1 0x'+hex_str[0:4])
    asm.append('ori $k1 $k1 0x'+hex_str[4:8])
    asm.append('sw $k1 '+hex(counter)+' $k0')
    counter = counter + 0x04
  return(asm)
    